/*
 * mbrola.c - SubAloud subtitles reader (common code for mbrola)
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>
#include <glib.h>

static void dtoa(double dt,char *buf)
{
    int n=(int)(100.0*dt);
    sprintf(buf,"%d.%02d",n/100,n%100);
}
    

short *mbrola_get_wave(char *voice,char *phonemes,double pitch,double rate,int *len,int *offset,int *freq)
{
    int pid;
    char *est[10],b1[10],b2[10];
    int ns;
    int from_mbrola[2];
    int to_mbrola[2];
    char *wave;
    int wave_length,wave_size;
    char bf[8192];
    
    if (pipe(from_mbrola)) {
	return NULL;
    }
    if (pipe(to_mbrola)) {
	close(from_mbrola[0]);
	close(from_mbrola[1]);
	return NULL;
    }
    pid=fork();
    if (pid < 0) {
	close(from_mbrola[0]);
	close(from_mbrola[1]);
	close(to_mbrola[0]);
	close(to_mbrola[1]);
	return NULL;
    }
    if (!pid) {
	ns=0;
	est[ns++]="mbrola";
	est[ns++]="-e";
	dtoa(pitch,b1);
	est[ns++]="-f";
	est[ns++]=b1;
	dtoa(rate,b2);
	est[ns++]="-t";
	est[ns++]=b2;
	est[ns++]=voice;
	est[ns++]="-";
	est[ns++]="-.wav";
	est[ns++]=NULL;
	close(from_mbrola[0]);
	close(to_mbrola[1]);
	dup2(from_mbrola[1],1);
	dup2(to_mbrola[0],0);
	execvp("mbrola",est);
	exit(1);
    }
    close(from_mbrola[1]);
    close(to_mbrola[0]);
    write(to_mbrola[1],phonemes,strlen(phonemes));
    write(to_mbrola[1],"\n",1);
    close(to_mbrola[1]);
    wave_size=65536;
    wave_length=0;
    wave=malloc(wave_size);
    for (;;) {
	int n=read(from_mbrola[0],bf,8192);
	if (n <= 0) break;
	while (n + wave_length > wave_size) {
	    wave_size=2*wave_size;
	    wave=realloc(wave,wave_size);
	}
	memcpy(wave+wave_length,bf,n);
	wave_length += n;
    }
    close(from_mbrola[0]);
    int status;
    waitpid(pid,&status,0);
    if (!wave_length) {
	free(wave);
	return NULL;
    }
    *len=wave_length/2-22;
    *offset=22;
    unsigned long *ws=(unsigned long *) wave;
    *freq=GULONG_FROM_LE(ws[6]);
    return (short *)wave;
    
}
