/*
 * modmilena.c - SubAloud subtitles reader (milena/ivolektor module)
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <milena.h>
#include <milena_mbrola.h>
#include <libivolektor.h>
#include <glib.h>

#include "../module.h"

int ivol_mode;
char *ivol_voice;
struct milena *milena;
struct milena_mbrola_cfg *mimbrola;
static struct IVOL_data *ivona;

static void find_mbrola_voice(void)
{
	static char *vpaths[]={
		"/usr/share/mbrola/voices/pl1",
		"/usr/share/mbrola/pl1/pl1",
		NULL};
	int i;
	for (i=0;vpaths[i];i++) {
		if (g_file_test(vpaths[i],G_FILE_TEST_IS_REGULAR)) {
			strcpy(mbrola_voice,vpaths[i]);
			break;
		}
	}
}


static void close_milena_libs(void)
{
	if (milena) milena_Close(milena);
	if (mimbrola) milena_CloseModMbrola(mimbrola);
	mimbrola=NULL;
	milena=NULL;
}

static int init_milena_libs(char **error)
{
	char ibuf[1024],obuf[1024],ixbuf[1024];
	int phmode;
	int i;
	struct miltheme *mt;
	if (!ivol_mode) {
		if (!mbrola_voice[0]) find_mbrola_voice();
		if (!mbrola_voice[0]) {
			*error=strdup("Podaj położenie pliku pl1 dla Mbroli");
			return 0;
		}
		if (!g_file_test(mbrola_voice,G_FILE_TEST_IS_REGULAR)) {
			*error=strdup("Podany plik pl1 nie istnieje.\nPodaj położenie pliku pl1 dla Mbroli\n");
			return 0;
		}
	}
	phmode=MILENA_PHR_IGNORE_INFO | MILENA_PHR_IGNORE_TILDE;
	close_milena_libs();
	milena=milena_Init(
		milena_FilePath(ivol_mode?"ive_pho.dat":"pl_pho.dat",obuf),
		milena_FilePath("pl_dict.dat",ibuf),
		milena_FilePath("pl_stress.dat",ixbuf));
	if (!milena) {
		*error=strdup("milena_init failed\n");
		return 0;
	}
	if (!milena_ReadPhraser(milena,
		milena_FilePath("pl_phraser.dat",ibuf))) {
			milena_Close(milena);
			milena=NULL;
			*error=strdup("milena_READPHRASER failed\n");
			return 0;
	}
	if (!milena_ReadUserDicWithFlags(milena,
		milena_FilePath("pl_udict.dat",ibuf),0)) {
			milena_Close(milena);
			milena=NULL;
			*error=strdup("milena_userdic failed\n");
			return 0;
	}
	if (ivol_mode) {
		milena_ReadUserDicWithFlags(milena,
			milena_FilePath("ive_udict.dat",ibuf),0);
		milena_ReadIvonaFin(milena,
			milena_FilePath("ive_fin.dat",ibuf));
	}
	if (!ivol_mode) {
		mimbrola=milena_InitModMbrola(
			milena_FilePath("pl_mbrola.dat",ibuf));
		if (!mimbrola) {
			close_milena_libs();
			*error=strdup("milena_initmodmbrola failed\n");
			return 0;
		}
		if (!milena_ReadInton(mimbrola,
			milena_FilePath("pl_intona.dat",ibuf))) {
			close_milena_libs();
			*error=strdup("milena_readinton failed\n");
			return 0;
		}
		if (!milena_ModMbrolaExtras(mimbrola,milena_FilePath("pl_pro_mbrola.dat",ibuf))) {
			close_milena_libs();
			*error=strdup("milena_mbrola_extras failed\n");

			return 0;
		}
		milena_ModMbrolaSetFlag(
				mimbrola,-1,MILENA_MBFL_BOOKMODE);
	}
	if (!milena_ReadPhraser(milena,	milena_FilePath("pl_hours.dat",ibuf))) {
		close_milena_libs();
		*error=strdup("milena read hours failed\n");
		return 0;
	}
	if (languages) {
	    char *cl;
	    for (cl=languages;*cl;) {
		char buf[4];
		while (*cl && (isspace(*cl) || *cl == ',')) cl++;
		if (!*cl) break;
		for (i=0;cl[i] && cl[i] != ',' && !isspace(cl[i]);i++) {
		    if (i < 2) buf[i]=cl[i];
		}
		cl += i;
		if (i<2) continue;
		buf[2]=0;
		sprintf(ibuf,"pl_%s_udic.dat",buf);
		milena_FilePath(ibuf,obuf);
		if (g_file_test(obuf,G_FILE_TEST_IS_REGULAR)) {
			if (!milena_ReadUserDic(milena,obuf)) {
				close_milena_libs();
				*error=strdup("milena_userdic_lang failed\n");

				return 0;
			}
		}
		sprintf(ibuf,"pl_%s_stress.dat",buf);
		milena_FilePath(ibuf,obuf);
		if (g_file_test(obuf,G_FILE_TEST_IS_REGULAR)) {
			if (!milena_ReadStressFile(milena,obuf)) {
				close_milena_libs();
				*error=strdup("milena_stress_lang failed\n");
				return 0;
			}
		}
		milena_SetLangMode(milena,buf);
	    }
	}
	if (themes) {
	    char *cl;
	    for (cl=themes;*cl;) {
		char buf[16],bth[32];
		while (*cl && (isspace(*cl) || *cl == ',')) cl++;
		if (!*cl) break;
		for (i=0;cl[i] && cl[i] != ',' && !isspace(cl[i]);i++) {
		    if (i < 15) buf[i]=cl[i];
		}
		cl += i;
		if (i<2) continue;
		if (i>15) continue;
		buf[i]=0;
		sprintf(bth,"pl_%s_theme.dat",buf);
		if (!milena_ReadPhraser(milena,milena_FilePath(bth,ibuf))) {
		    fprintf(stderr,"No such theme: [%s]\n",buf);
		}
	    }
	}

	if (dictfile) {
	    if (!milena_ReadUserDicWithFlags(milena,dictfile,MILENA_UDIC_IGNORE)) {
		fprintf(stderr,"Cannot read dictionary: %s\n",dictfile);
	    }
	}
	if (phmode & 3) milena_SetPhraserMode(milena,phmode & 3);
	milena_ReadVerbs(milena,milena_FilePath("pl_verbs.dat",ibuf));
	return 1;
}



static char *_init_ivona(char **err)
{
    if (!init_milena_libs(err)) {
	return 0;
    }
    ivona=IVOL_open(ivol_voice,err);
    if (!ivona) {
	close_milena_libs();
        return NULL;
    }
    return strdup("trans=orto\njoin\nwave\n");
}

static char *_init_mbrola(char **err)
{
    if (!mbrola_pitch_set) mbrola_pitch=0.9;
    if (!init_milena_libs(err)) return NULL;
    return strdup("trans=mbrola\njoin\nwave\nrwave\n");
}

char *init_speaker_module(char **error)
{
    int scmp(char *s1,char *s2,int n)
    {
	if (strlen(s2) != n) return 0;
	return !strncmp(s1,s2,n);
    }
    if (!module_voice) {
	ivol_mode=0;
	ivol_voice=NULL;
    }
    else {
	if (!strcasecmp(module_voice,"Milena")) {
	    ivol_mode=0;
	    ivol_voice=NULL;
	}
	else {
	    ivol_mode=1;
	    ivol_voice=module_voice;
	}
    }
    if (ivol_mode) {
	return _init_ivona(error);
    }
    else {
	return _init_mbrola(error);
    }
}
/* mbrola part */

static int count_syllables(char *str)
{
	int syl=0;
	for (;*str;str++) {
		if (strchr("aeiouyYMOE",*str)) syl++;
	}
	return syl;
}

static void translate_phrase(char *phrase,char **otbuf,int *blen)
{
	static char *buf1,*buf2;
	static int len1,len2;
	int n;
	if (!len1) {
		buf1=g_malloc(len1=8192);
	}
	if (!len2) {
		buf2=g_malloc(len2=8192);
	}
	n=milena_Prestresser(milena,phrase,buf1,len1);
	if (n) {
		len1=n+1024;
		buf1=g_realloc(buf1,len1);
		milena_Prestresser(milena,phrase,buf1,len1);
	}
	n=milena_TranslatePhrase(milena,(unsigned char *)buf1,buf2,len2,0);
	if (n) {
		len2=n+1024;
		buf2=g_realloc(buf2,len2);
		milena_TranslatePhrase(milena,(unsigned char *)buf1,buf2,len2,0);
	}
	n=milena_Poststresser(buf2,*otbuf,*blen);
	if (n) {
		*blen = n+1024;
		*otbuf=g_realloc(*otbuf,*blen);
		milena_Poststresser(buf2,*otbuf,*blen);
	}
	
}


static char *mil_translate(char *par)
{
	static char *phrase_buf,*sentence_buf,*second_buf;
	static int phrase_len,sentence_len,second_len;
	int this_ptyp,next_ptyp;
	int have_phrase=0;
	static struct phone_buffer phb;
	if (!phb.str) {
	    phb.str=malloc(65536);
	    phb.buf_len=65536;
	}
	phb.str_len=0;
	*phb.str=0;
	int get_phrase()
	{
		int nlen;
		char *opar=par;
		if (have_phrase) {
			char *c=sentence_buf;
			int n=sentence_len;
			sentence_buf=second_buf;
			second_buf=c;
			sentence_len=second_len;
			second_len=n;
			have_phrase=0;
			this_ptyp=next_ptyp;
			return 1;
		}

		nlen=milena_GetPhrase(milena,&par,phrase_buf,phrase_len,&this_ptyp);
		if (nlen<0) return 0;
		if (nlen) {
			par=opar;
			phrase_len=nlen+1024;
			phrase_buf=g_realloc(phrase_buf,phrase_len);
			milena_GetPhrase(milena,&par,phrase_buf,phrase_len,&this_ptyp);
		}
		translate_phrase(phrase_buf,&sentence_buf,&sentence_len);
		return 1;
	}

	int get_next_phrase()
	{
		int nlen;
		char *c=par;
		nlen=milena_GetPhrase(milena,&par,phrase_buf,phrase_len,&next_ptyp);
		if (nlen<0) return 0;
		if (nlen) {
			phrase_len=nlen+1024;
			par=c;
			phrase_buf=g_realloc(phrase_buf,phrase_len);
			milena_GetPhrase(milena,&par,phrase_buf,phrase_len,&next_ptyp);
		}
		translate_phrase(phrase_buf,&second_buf,&second_len);
		have_phrase=1;
		return 1;

	}
	if (!phrase_len) {
		phrase_buf=g_malloc(phrase_len=8192);
		sentence_buf=g_malloc(sentence_len=8192);
		second_buf=g_malloc(second_len=8192);
	}
	for (;;) {
		if (!get_phrase()) {
		    break;
		}

		if ((this_ptyp & 7) == 1) {
			if (!get_next_phrase()) this_ptyp=(this_ptyp & 0xf8) | 5;
			else {
				int b=0;
				if ((next_ptyp & 7) == 2) {
					if (count_syllables(second_buf)<6) b=1;
				}
				//if (b) this_ptyp=(this_ptyp & 0xf8) | 4;
				if (b) this_ptyp |= 128;
			}

		}
		if ((this_ptyp & 7)==7) {
			this_ptyp=(this_ptyp & 0xf8) | 4;
		}
		milena_ModMbrolaGenPhraseP(mimbrola,sentence_buf,&phb,this_ptyp);
	}
	if (phb.str_len + 20 > phb.buf_len) {
	    phb.buf_len *= 2;
	    phb.str=realloc(phb.str,phb.buf_len);
	}
	strcat(phb.str,"_ 2\n");
	return g_strdup(phb.str);
}


/* IVONA part */

static void milena_ivona_push(int mode,char *inbuf,char **outbuf,int *outsize)
{
    int len;
    len=milena_ivonizer_n(mode,inbuf,*outbuf,*outsize,milena);
    if (len) {
        *outsize=len+256;
        if (*outbuf) *outbuf=realloc(*outbuf,*outsize);
        else {
            *outbuf=g_malloc(*outsize);
            **outbuf=0;
        }
        milena_ivonizer_n(mode,inbuf,*outbuf,*outsize,milena);
    }
}

static void ivona_phrase(char *phrase,char **outbuf,int *blen)
{
	static char *buf1;
	static int len1;
	int n;
	if (!len1) {
		buf1=g_malloc(len1=8192);
	}
	n=milena_Prestresser(milena,phrase,buf1,len1);
	if (n) {
		len1=n+1024;
		buf1=g_realloc(buf1,len1);
		milena_Prestresser(milena,phrase,buf1,len1);
	}
	n=milena_TranslatePhrase(milena,(unsigned char *)buf1,*outbuf,*blen,0);
	if (n) {
		*blen=n+1024;
		*outbuf=g_realloc(*outbuf,*blen);
		milena_TranslatePhrase(milena,(unsigned char *)buf1,*outbuf,*blen,0);
	}
}


static char *ivo_translate(char *iso2str)
{
	static char *phrase_buf,*trans_buf;
	static int phrase_len,trans_len;
	char *ivona_outbuf=NULL;
	int ivona_outsize=0;
	if (!phrase_buf) phrase_buf=malloc(phrase_len=1024);
	if (!trans_buf) trans_buf=malloc(trans_len=1024);
	for (;;) {
   		int nlen,ptyp;
		nlen=milena_GetPhrase(milena,&iso2str,phrase_buf,phrase_len,&ptyp);
		if (nlen<0) break;
		if (nlen) {
			phrase_len=nlen+1024;
			phrase_buf=realloc(phrase_buf,phrase_len);
			milena_GetPhrase(milena,&iso2str,phrase_buf,phrase_len,&ptyp);
		}
		ivona_phrase(phrase_buf,&trans_buf,&trans_len);
		milena_ivona_push(ptyp & 7,trans_buf,&ivona_outbuf,&ivona_outsize);
	}
	return ivona_outbuf;
}

short *get_timed_wave(char *str,int *len,int *offset,int *freq,double rate)
{
    if (ivol_mode) return NULL;
    return mbrola_get_wave(mbrola_voice,str,mbrola_pitch,1.0/rate,len,offset,freq);
}


short *get_wave(char *str,int *len,int *offset,int *freq)
{
	short *wave;
	int length,offs,rest;
	char *err;
	if (!str || !*str) return NULL;
	if (!ivol_mode) {
	    //*freq=16000;
	    //*offset=0;
	    return mbrola_get_wave(mbrola_voice,str,mbrola_pitch,1.0,len,offset,freq);
	}
        wave=IVOL_getWave(ivona,str,&length,&err);
        if (!wave) {
	    fprintf(stderr,"ERR %s\n",err);
	    return NULL;
	}
        short *wav2=IVOL_Repauser(ivona,wave,&length,0.25,&offs,&rest);
        if (wav2) {
            free(wave);
            wave=wav2;
        }
	*len=length-(rest+offs);
	*offset=offs;
	*freq=ivona->freq;
	return wave;
}

char *get_phones(char *str)
{
    int sajz=milena_utf2iso(str,NULL,1,NULL);
    char *cs=g_malloc(sajz+1);
    char *rc;
    milena_utf2iso(str,cs,1,NULL);
    if (!ivol_mode) {
	rc=mil_translate(cs);
    }
    else {
	rc=ivo_translate(cs);
    }
    g_free(cs);
    return rc;
}
