/*
 * module_main.c - SubAloud subtitles reader (common module code)
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <ctype.h>
#include "module.h"
#include <getopt.h>


/*
Standard module parameters should be:

a, datapath - path to module private data
v, mbrola_voice - path to mbrola database file
p, mbrola_pitch - pitch for mbrola voices (0.9 for Milena, 1.0 for rest)
l, languages - comma separated list of foreign languages for NLP (may be ignored)
t, themes - comma separated list of themes for NLP (may be ignored)
d, dict - dictionary file for NLP 
*/

char *module_voice;
char mbrola_voice[256];
double mbrola_pitch=1.0;
int mbrola_pitch_set;
char datapath[256];
char *languages;
char *themes;
char *dictfile;
char *preproc;
int preproc_type;
 
char *trim(char *c)
{
    char *d,*e;
    while (*c && isspace(*c)) c++;
    for (d=e=c;*d;d++) if (!isspace(*d)) e=d+1;
    *e=0;
    return c;
}

char *get_command(char **param)
{
    char buf[16384],*rc,*pr;
    int prl,prs,pl;
    if (!fgets(buf,16384,stdin)) return NULL;
    *param=NULL;
    if (buf[0] != '@') {
	return g_strdup(trim(buf));
    }
    rc=g_strdup(trim(buf+1));
    prs=2048;
    pr=g_malloc(prs);
    prl=0;
    
    for (;;) {
	if (!fgets(buf,16384,stdin)) {
	    g_free(rc);
	    return NULL;
	}
	char *c=buf;
	if (*buf=='.') {
	    if (buf[1]=='\r' || buf[1] == '\n') break;
	    c++;
	}
	pl=strlen(c);
	if (pl+prl >= prs-1) {
	    prs = 2*(pl+prl);
	    pr=g_realloc(pr,prs);
	}
	memcpy(pr+prl,c,pl);
	prl += pl;
    }
    pr[prl]=0;
    *param=pr;
    return rc;
}


void _emit_out(int nr,char *str)
{
    char *c;
    for (;;) {
	c=strchr(str,'\n');
	if (!c) {
	    printf("%d %s\n",nr,str);
	    break;
	}
	c++;
	if (!*c) {
	    printf("%d %s",nr,str);
	    break;
	}
	printf("%d-%-*.*s",nr,c-str,c-str,str);
	str=c;
    }
    fflush(stdout);
}

void send_wave(int freq,short *wave,int len)
{
    printf("300-WAVE DATA FOLLOWS:%d:%d\n",freq,len);
    fwrite(wave,2,len,stdout);
    printf("\n300 OK\n");
    fflush(stdout);
}

char *_get_param(char **str)
{
    char *rc;
    while (**str && isspace(**str)) (*str)++;
    if (!**str) return NULL;
    rc=*str;
    while (**str && !isspace(**str)) (*str)++;
    if (**str) {
	*(*str)++ = 0;
	while (**str && isspace(**str)) (*str)++;
    }
    return rc;
}

    
int do_command(char *cmdp,char *par)
{
    char *cmd;
    if (!*cmdp) return 0;
    cmd=_get_param(&cmdp);
    if (!strcasecmp(cmd,"quit")) {
	printf("299 GOOD BYE\n");
	fflush(stdout);
	return 1;
    }
    if (!strcasecmp(cmd,"trans")) {
	char *rc=get_phones(cmdp);
	if (!rc) {
	    printf("500 INTERNAL ERROR\n");
	    fflush(stdout);
	    return 0;
	}
	_emit_out(201,rc);
	g_free(rc);
	return 0;
    }
    if (!strcasecmp(cmd,"wave")) {
	int len,offset,freq;
	short *wave;
	if (!*cmdp) cmdp=par;
	wave=get_wave(cmdp,&len,&offset,&freq);
	if (!wave) {
	    printf("500 INTERNAL ERROR\n");
	    fflush(stdout);
	    return 0;
	}
	send_wave(freq,wave+offset,len);
	free(wave);
	return 0;
    }
    if (!strcasecmp(cmd,"rwave")) {
	char *cs=_get_param(&cmdp);
	int len,offset,freq;
	short *wave;
	if (!cs) {
	    printf("400 REQUEST ERROR\n");
	    fflush(stdout);
	    return 0;
	}
	if (!*cmdp) cmdp=par;
	wave=get_timed_wave(cmdp,&len,&offset,&freq,strtod(cs,NULL));
	if (!wave) {
	    printf("500 INTERNAL ERROR\n");
	    fflush(stdout);
	    return 0;
	}
	send_wave(freq,wave+offset,len);
	free(wave);
	return 0;
    }
    printf("550 UNKNOWN COMMAND '%s'\n",cmd);
    fflush(stdout);
    return 0;
}

void erxt(char *s)
{
    _emit_out(599,s);
    exit(1);
}

int main(int argc,char *argv[])
{
    char *cmd,*par,*err;int rc;
    char *optstr="a:v:p:l:t:d:P:";
    struct option options[]={
	{"datapath",1,NULL,'a'},
	{"mbrola-voice",1,NULL,'v'},
	{"mbrola-pitch",1,NULL,'p'},
	{"languages",1,NULL,'l'},
	{"themes",1,NULL,'t'},
	{"dict",1,NULL,'d'},
	{"pre",1,NULL,'P'},
	{NULL,0,NULL,0}};

    /*
     * we NEED glib memory allocation to be compatible with malloc
     */
    static GMemVTable GMV={
	malloc,realloc,free,
	NULL,NULL,NULL}; 
    g_mem_set_vtable(&GMV);
    
    for (;;) {
	int opt=getopt_long(argc,argv,optstr,options,NULL);
	if (opt < 0) break;
	switch(opt) {
	    
	    case 'a':
	    if (strlen(optarg) < 253) {
		strcpy(datapath,optarg);
	    }
	    else {
		erxt("Data path too long");
	    }
	    break;
	    
	    case 'v':
	    if (strlen(optarg) < 253) {
		strcpy(mbrola_voice,optarg);
	    }
	    else {
		erxt("Mbrola path too long");
	    }
	    break;


	    case 'p':
	    mbrola_pitch=strtod(optarg,NULL);
	    if (mbrola_pitch < 0.5 || mbrola_pitch > 2.0) {
		erxt("Mbrola pitch out of range");
	    }
	    mbrola_pitch_set=1;
	    break;
	    
	    case 'P':
	    if (!strcmp(optarg,"standard")) {
		preproc_type = PREPROC_STANDARD;
	    }
	    else if (!strncmp(optarg,"standard:",9)) {
		preproc_type = PREPROC_STANDARD;
		preproc=strdup(optarg+9);
	    }
	    else if (!strcmp(optarg,"preproc")) {
		preproc_type=PREPROC_PREPROC;
	    }
	    else if (!strncmp(optarg,"preproc:",8)) {
		preproc_type = PREPROC_STANDARD;
		preproc=strdup(optarg+8);
	    }
	    else if (!strncmp(optarg,"extern:",7)) {
		preproc_type = PREPROC_EXTERNAL;
		preproc=strdup(optarg+7);
	    }
	    else if (!strcmp(optarg,"none")) {
		preproc_type = PREPROC_NONE;
	    }
	    else {
		preproc_type = PREPROC_PRIVATE;
		preproc=strdup(optarg);
	    }
	    break;

	    case 'l':
	    languages=strdup(optarg);
	    break;

	    case 't':
	    themes=strdup(optarg);
	    break;
	    
	    case 'd':
	    dictfile=strdup(optarg);
	    break;
	    
	    default:
	    erxt("Unknown module parameter");
	}
    }
    if (!preproc_type) preproc_type=PREPROC_STANDARD;
    if (optind < argc) {
	module_voice=strdup(argv[optind++]);
    }
    if (!(cmd=init_speaker_module(&err))) {
	if (!err) err="INTERNAL ERROR";
	_emit_out(501,err);
	exit(1);
    }
    _emit_out(200,cmd);
    g_free(cmd);
    for (;;) {
	par=NULL;
	cmd=get_command(&par);
	if (!cmd) break;
	rc=do_command(cmd,par);
	g_free(cmd);
	if (par) g_free(par);
	if (rc) break;
    }
    exit(0);
}
