
#include <glib.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include "module.h"

static char *preproc_lst,*preproc_abk;

int init_preproc(char **error)
{
    char buf[256];
    char *c;
    static char *prepdir[]={
	"/usr/share/preproc/",
	"/usr/local/share/preproc/",
	"/usr/local/preproc",
	NULL};
    if (!preproc) {
	int i;
	for (i=0;prepdir[i];i++) {
	    sprintf(buf,"%sRules.lst",prepdir[i]);
	    if (g_file_test(buf,G_FILE_TEST_IS_REGULAR)) {
		preproc_lst=g_strdup(buf);
		sprintf(buf,"%sHadifix.abk",prepdir[i]);
		if (g_file_test(buf,G_FILE_TEST_IS_REGULAR)) {
		    preproc_abk=g_strdup(buf);
		}
		return 1;
	    }
	}
	*error=g_strdup("Preproc can't find Rules.lst");
	return 0;
    }
    c=strchr(preproc,':');
    if (c) {
	preproc_lst=g_strndup(preproc,c-preproc);
	preproc_abk=g_strdup(c+1);
    }
    else {
	preproc_lst=g_strdup(preproc);
    }
    if (!g_file_test(preproc_lst,G_FILE_TEST_IS_REGULAR)) {
	*error=g_strdup_printf("Can't find rules file '%s'",preproc_lst);
	return 0;
    }
    if (preproc_abk && !g_file_test(preproc_abk,G_FILE_TEST_IS_REGULAR)) {
	*error=g_strdup_printf("Can't find abbreviations file '%s'",preproc_abk);
	return 0;
    }
    return 1;
}

char *preproc_string(char *str,int recode_in, int recode_out)
{
    char bf[8192];
    int from_preproc[2],to_preproc[2];
    char *cvt=NULL;
    if (recode_in) {
	GIConv gi=g_iconv_open("ISO-8859-2//TRANSLIT","UTF-8");
	if (!gi) return NULL;
	cvt=g_convert_with_iconv(str,-1,gi,NULL,NULL,NULL);
	if (!cvt) return NULL;
	g_iconv_close(gi);
    }
    if (pipe(from_preproc)) {
	if (cvt) g_free(cvt);
	return NULL;
    }
    if (pipe(to_preproc)) {
	if (cvt) g_free(cvt);
	close(from_preproc[0]);
	close(from_preproc[1]);
	return NULL;
    }
    int pid=fork();
    if (pid < 0) {
	if (cvt) g_free(cvt);
	close(from_preproc[0]);
	close(from_preproc[1]);
	close(to_preproc[0]);
	close(to_preproc[1]);
	return NULL;
    }
    if (!pid) {
	char *args[10];
	int n;
	n=0;
	args[n++]="preproc";
	args[n++]=preproc_lst;
	if (preproc_abk) {
	    args[n++]=preproc_abk;
	}
	args[n]=NULL;
	close(from_preproc[0]);
	close(to_preproc[1]);
	dup2(from_preproc[1],1);
	dup2(to_preproc[0],0);
	execvp("preproc",args);
	exit(1);
    }
    close(from_preproc[1]);
    close(to_preproc[0]);
    if (cvt) {
	write(to_preproc[1],cvt,strlen(cvt));
	g_free(cvt);
    } else {
	write(to_preproc[1],str,strlen(str));
    }
    write(to_preproc[1],"\n",1);
    close(to_preproc[1]);
    int out_size=65536;
    int out_length=0;
    char *outstr=g_malloc(out_size);
    for (;;) {
	int n=read(from_preproc[0],bf,8192);
	if (n <= 0) break;
	while (n + out_length >= out_size) {
	    out_size=2*out_size;
	    outstr=g_realloc(outstr,out_size);
	}
	memcpy(outstr+out_length,bf,n);
	out_length += n;
    }
    close(from_preproc[0]);
    int status;
    waitpid(pid,&status,0);
    if (!out_length) {
	g_free(outstr);
	return NULL;
    }
    outstr[out_length]=0;
    if (recode_out) {
	GIConv gi=g_iconv_open("UTF-8","ISO-8859-1");
	if (!gi) return NULL;
	cvt=g_convert_with_iconv(str,-1,gi,NULL,NULL,NULL);
	if (!cvt) return NULL;
	g_iconv_close(gi);
    }
    return outstr;
}
