/*
 * modsapi.c - SubAloud subtitles reader (sapilektor module)
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <libsapilektor.h>

#include "../module.h"

static struct SAPI_data *lector;

char *init_speaker_module(char **error)
{
    if (!module_voice) {
	*error=strdup("Module voice not set");
	return NULL;
    }
    if (preproc_type == PREPROC_PREPROC) {
	if (!init_preproc(error)) {
	    return NULL;
	}
    }
    else if (preproc_type != PREPROC_NONE) {
	preproc_type = PREPROC_STANDARD;
    }
    lector=SAPI_open(module_voice,error);
    if (!lector) {
        return NULL;
    }
    return strdup("trans=private\nwave\n");
}

char *get_phones(char *str)
{
    char *c;
    gunichar z;
    int fnd=0;
    if (strstr(str,">>>") || strstr(str,"<<<")) return NULL;
    if (preproc_type == PREPROC_STANDARD) return simplify_text(str);
    for (c=str;*c;) {
	z=g_utf8_get_char(c);
	if (g_unichar_isalnum(z)) {
	    fnd=1;
	    break;
	}
	c=g_utf8_next_char(c);
    }
    if (!fnd) return NULL;
    return g_strdup(str);
}

short *get_timed_wave(char *str,int *len,int *offset,int *freq,double rate)
{
    return NULL;
}


short *get_wave(char *str,int *len,int *offset,int *freq)
{
	short *wave;
	int length,offs,rest;
	char *err;
	if (!str || !*str) return NULL;
	if (preproc_type == PREPROC_PREPROC) {
	    char *s=preproc_string(str,1,1);
	    if (!s) return NULL;
	    wave=SAPI_getWave(lector,s,&length,&err);
	    g_free(s);
	}
	else {
	    wave=SAPI_getWave(lector,str,&length,&err);
	}
        if (!wave) {
	    fprintf(stderr,"ERR %s\n",err);
	    return NULL;
	}
        short *wav2=SAPI_Repauser(lector,wave,&length,0.25,&offs,&rest);
        if (wav2) {
            free(wave);
            wave=wav2;
        }
	*len=length-(rest+offs);
	*offset=offs;
	*freq=lector->freq;
	return wave;
}
