/*
 * modtxt2pho.c - SubAloud subtitles reader (txt2pho module)
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "../module.h"

char *voice_sex;
int voice_freq;
double MBROLA_PITCH=1.0;


static char *find_voice_sex(void)
{
	static char *sexy[]={NULL,NULL,"-m",NULL,"-m","-f","-m","-f"};
	static int freqs[]={0,0,16000,0,16000,22050,22050,22050};
	if (strlen(module_voice)!=3) return NULL;
	if (strncmp(module_voice,"de",2)) return;
	int i=module_voice[2]-'0';
	if (i < 2 || i > 7 || i==3) return;
	voice_sex=sexy[i];
	voice_freq=freqs[i];
	return voice_sex;
}

static void find_mbrola_voice(void)
{
	static char *vpaths[]={
		"/usr/share/mbrola/voices/%s",
		"/usr/share/mbrola/%s/%s",
		NULL};
	char buf[256];
	int i;
	if (strlen(module_voice)!=3) return;
	if (strncmp(module_voice,"de",2)) return;
	i=module_voice[2]-'0';
	if (i < 2 || i > 7 || i==3) return;
	for (i=0;vpaths[i];i++) {
		sprintf(buf,vpaths[i],module_voice,module_voice);
		if (g_file_test(buf,G_FILE_TEST_IS_REGULAR)) {
			strcpy(mbrola_voice,buf);
			break;
		}
	}
}

static void find_data_path(void)
{
    static char *datas[]={
	"/usr/local/txt2pho/data/",
	"/usr/local/share/txt2pho/data/",
	"/usr/share/txt2pho/data/",
	NULL};
    int i;
    for (i=0;datas[i];i++) {
	if (g_file_test(datas[i],G_FILE_TEST_IS_DIR)) {
	    strcpy(datapath,datas[i]);
	    break;
	}
    }
}


char *init_speaker_module(char **error)
{
    if (!module_voice) {
	*error=strdup("Module voice not set");
	return NULL;
    }
    if (!find_voice_sex()) {
	*error=strdup("Bad voice");
	return NULL;
    }
    if (!mbrola_voice[0]) find_mbrola_voice();
    if (!mbrola_voice[0]) {
	*error=strdup("Voice not found");
	return NULL;
    }
    if (!datapath[0]) {
	find_data_path();
	if (!datapath[0]) {
	    *error=strdup("Txt2pho data path not found");
	    return NULL;
	}
    }
    else {
	if (datapath[strlen(datapath)-1] != '/') {
	    strcat(datapath,"/");
	}
    }
    if (preproc_type == PREPROC_PREPROC) {
	if (!init_preproc(error)) {
	    return NULL;
	}
    }
    else if (preproc_type != PREPROC_NONE) {
	preproc_type = PREPROC_STANDARD;
    }
    return strdup("trans=mbrola\njoin\nwave\nrwave\n");
}

char *txt2pho(char *cvt)
{
    char bf[8192];
    int from_t2f[2],to_t2f[2];
    if (pipe(from_t2f)) {
	return NULL;
    }
    if (pipe(to_t2f)) {
	close(from_t2f[0]);
	close(from_t2f[1]);
	return NULL;
    }
    int pid=fork();
    if (pid < 0) {
	close(from_t2f[0]);
	close(from_t2f[1]);
	close(to_t2f[0]);
	close(to_t2f[1]);
	return NULL;
    }
    if (!pid) {
	char *args[10];
	int n;
	n=0;
	args[n++]="txt2pho";
	args[n++]="-p";
	args[n++]=datapath;
	args[n++]=voice_sex;
	args[n]=NULL;
	close(from_t2f[0]);
	close(to_t2f[1]);
	dup2(from_t2f[1],1);
	dup2(to_t2f[0],0);
	execvp("txt2pho",args);
	exit(1);
    }
    close(from_t2f[1]);
    close(to_t2f[0]);
    write(to_t2f[1],cvt,strlen(cvt));
    write(to_t2f[1],"\n",1);
    close(to_t2f[1]);
    int phone_size=65536;
    int phone_length=0;
    char *phones=g_malloc(phone_size);
    for (;;) {
	int n=read(from_t2f[0],bf,8192);
	if (n <= 0) break;
	while (n + phone_length >= phone_size) {
	    phone_size=2*phone_size;
	    phones=g_realloc(phones,phone_size);
	}
	memcpy(phones+phone_length,bf,n);
	phone_length += n;
    }
    close(from_t2f[0]);
    int status;
    waitpid(pid,&status,0);
    if (!phone_length) {
	g_free(phones);
	return NULL;
    }
    phones[phone_length]=0;
    return phones;
}


short *get_timed_wave(char *str,int *len,int *offset,int *freq,double rate)
{
	short *wave;
	int length,offs,rest;
	char *err,*pstr;
	if (!str || !*str) return NULL;
	if (preproc_type == PREPROC_PREPROC) {
	    char *s=preproc_string(str,0,0);
	    if (!s) return NULL;
	    pstr=txt2pho(s);
	    g_free(s);
	}
	else {
	    pstr=txt2pho(str);
	}
	if (!pstr) return NULL;
	if (!*str) {
	    g_free(pstr);
	    return NULL;
	}
	//*freq=voice_freq;
	//*offset=0;
	wave=mbrola_get_wave(mbrola_voice,pstr,mbrola_pitch,1.0/rate,len,offset,freq);
	g_free(pstr);
	return wave;
}


short *get_wave(char *str,int *len,int *offset,int *freq)
{
    	return get_timed_wave(str,len,offset,freq,1.0);
}


char *get_phones(char *str)
{
    char *s=NULL;
    if (preproc_type == PREPROC_STANDARD) {
	s=simplify_text(str);
	if (!s) return NULL;
    }
    else {
	// for NONE or PREPROC
	char *c;gunichar z;
	int fnd=0;
	for (c=str;*c;) {
	    z=g_utf8_get_char(c);
	    if (g_unichar_isalnum(z)) {
		fnd=1;
		break;
	    }
	    c=g_utf8_next_char(c);
	}
	if (!fnd) return NULL;
    }
    GIConv gi=g_iconv_open("ISO-8859-1//TRANSLIT","UTF-8");
    if (!gi) return NULL;
    char *cvt=g_convert_with_iconv(s?s:str,-1,gi,NULL,NULL,NULL);
    g_iconv_close(gi);
    if (s) g_free(s);
    return cvt;
}
