/*
 * mplrun.c - SubAloud subtitles reader
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include "videosa.h"

int mplayer_stdin;
int mplayer_pass;

static int mplayer_fd;
int mplayer_pid;
int mplayer_status;
int current_decisec;
int fps_set;
double fps;
static int nlast;
static GThread *mplayer_Thread;

int readline(int fd,char *buf)
{
	int n;
	char z;
	struct timeval tv;
	fd_set inset;
	n=read(fd,&z,1);
	if (!n) {
	    return -2;
	}
	if (n < 0) {
		if (errno != EAGAIN) return -2;
		FD_ZERO(&inset);
		FD_SET(fd,&inset);
		tv.tv_sec=0;
		tv.tv_usec=200000;
		n=select(fd+1,&inset,NULL,NULL,&tv);
		if (n == -1) return -2;
		if (!n) return -1;
		n=read(fd,&z,1);
		if (n <= 0) return -1;
	}
	for (;;) {
		if (z == '\n') break;
		*buf++ = z;
		n=read(fd,&z,1);
		if (n != 1) break;
		if (mplayer_pass) write(1,&z,1);
	}
	*buf=0;
	return 0;
}

int lineclock(int fd)
{
	char buf[256],*cs;
	int n,cds;
	struct timeval ntv;
	static struct timeval ltv;
	n=readline(fd,buf);
	if (n == -2) {
		return 0;
	}
	if (n == -1) {
		mplayer_status=VSFL_STOPPED;
		return 1;
	}
	if (!fps_set && !strncmp(buf,"ID_VIDEO_FPS=",13)) {
		fps=strtod(strchr(buf,'=')+1,NULL);
		fps_set=1;
		return 1;
	}
	cs=strstr(buf,"V:");
	if (!cs) return 1;
	cs+=2;
	while (*cs && isspace(*cs)) cs++;
	if (!isdigit(*cs)) return 1;
	cds=strtol(cs,&cs,10)*10;
	if (*cs++ == '.') {
		if (isdigit(*cs)) cds += (*cs) - '0';
	}
	current_decisec=cds;
	if (current_decisec < nlast) {
		mplayer_status=VSFL_BACKWARD;
		gettimeofday(&ltv,NULL);
	}
	else if (current_decisec > nlast + 1) {
		mplayer_status=VSFL_FORWARD;
		gettimeofday(&ltv,NULL);
	}
	else {
		if (mplayer_status & (VSFL_FORWARD | VSFL_BACKWARD)) {
			long usec;
			gettimeofday(&ntv,0);
			usec = ntv.tv_sec - ltv.tv_sec;
			usec= 1000000 * usec + (ntv.tv_usec + ltv.tv_usec);
			if (usec < 0 || usec > 1000000) {
				mplayer_status=VSFL_RUNNING;
			}
		}
		else {
			mplayer_status=VSFL_RUNNING;
		}
	}
	nlast=current_decisec;
	return 1;
}


gpointer mplayer_ThreadFunc(gpointer data)
{
    for(;;) {
	if (!lineclock(mplayer_fd)) {
	    mplayer_status=VSFL_FINISHED;
	    break;
	}
    }
    return NULL;
}

void start_mplayer(int argc,char *args[],int is_frames)
{
    int from_mplayer[2];
    long flags;
    if (!mplayer_stdin) {
	if(pipe(from_mplayer)<0) {
	    perror("pipe");
	    exit(1);
	}
	mplayer_pid=fork();
	if (mplayer_pid < 0) {
	    perror("fork");
	    exit(1);
	}
	if (!mplayer_pid) {
	    char **st=g_malloc(sizeof(*st) * (argc+10));
	    int i=0,j;
	    st[i++]="mplayer";
	    if (is_frames) st[i++]="-identify";
	    st[i++]="-noquiet";
	    for (j=0;j<argc;j++) st[i++]=args[j];
	    st[i++]=movie_name;
	    st[i]=NULL;
	    close(from_mplayer[0]);
	    dup2(from_mplayer[1],1);
	    fprintf(stderr,"Running mplayer\n");
	    execvp("mplayer",st);
	    exit(1);
	}
	close(0);
	close(from_mplayer[1]);
	mplayer_fd=from_mplayer[0];
    }
    else {
	mplayer_fd=0;
    }
    mplayer_status = VSFL_RUNNING;

    flags = fcntl(mplayer_fd,F_GETFL);
    fcntl(mplayer_fd,F_SETFL,flags | O_NONBLOCK);

#if !GLIB_CHECK_VERSION(2,32,0)
    g_thread_init(NULL);
    mplayer_Thread=g_thread_create(mplayer_ThreadFunc,NULL,TRUE,NULL);
#else
    mplayer_Thread=g_thread_new("vms_mplayer",mplayer_ThreadFunc,NULL);
#endif
}
