/*
 * subreader.c - SubAloud subtitles reader
 * Copyright (C) Bohdan R. Rau 2013 <ethanak@polip.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write see:
 *               <http://www.gnu.org/licenses/>.
 */
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <glib.h>
#include <stdio.h>
#include <enca.h>
#include "videosa.h"
#include <sys/stat.h>
#include <unistd.h>
#include <math.h>


struct subline {
	struct subline *next;
	int t1,t2;
	char *lines[8];
};

/*
static int msec(int t)
{
	return t % 1000;
}

static int sec(int t)
{
	return (t/1000) % 60;
}

static int mins(int t)
{
	return (t/60000) % 60;
}

static int hour(int t)
{
	return t/3600000;
}
*/

static int tester_pasuje(char *s,char *patr)
{
	char z,*c;
	while (*s && isspace(*s)) s++;
	while (*patr) {
		z=*patr++;
		if (!*s) return 0;
		if (z==' ') {
			if (!isspace(*s++)) return 0;
			while (*s && isspace(*s)) s++;
			continue;
		}
		if (z=='9') {
			if (!isdigit(*s++)) return 0;
			continue;
		}
		if (z=='5') {
			if (*s < '0' || *s > '5') return 0;
			s++;
			continue;
		}
		if (z=='d') {
			if (!isdigit(*s++)) return 0;
			while (*s && isdigit(*s)) s++;
			continue;
		}
		if (z=='D') {
			while (*s && isdigit(*s)) s++;
			continue;
		}
		if (z=='E') {
			c=s;
			while (*c && isdigit(*c)) c++;
			if (isspace(*c)) {
				s=c;
				while (*s && isspace(*s)) s++;
			}
			continue;
		}
		if (z=='S') {
			if (!isdigit(*s)) return 0;
			if (*s++ <= '5') {
				if (isdigit(*s)) s++;
			}
			continue;
		}
		if (z != *s++) return 0;
	}
	return 1;
}


// testery
// MPL2: [d][d]
// Micro: {d}{D}
// TMP: 99:59:59:
// SubRip: Ed:59:59,d --> d:59:59,d
// SubViewer: d:S:S:99,d:S:S:99
// JacoSUB: d:S:S.99 d:S:S.99


static struct subline *alloc_subline(char *src,int t1,int t2)
{
	struct subline *sr=g_malloc0(sizeof(struct subline));
	if (src) {
		sr->lines[0]=g_strdup(src);
	}
	sr->t1=t1;
	sr->t2=t2;
	return sr;
}

static struct subtitle *prepare_line(struct subline *sl,int unit_frames)
{
    int len,i;char *cs,*c,*d;
    struct subtitle *line;
    
    for (i=len=0;i<8 && sl->lines[i];i++) {
	c=d=sl->lines[i];
	if (*c=='/') c++;
	for (;*c;) {
	    if (*c != '{') {
		*d++=*c++;
		continue;
	    }
	    c=strchr(c,'}');
	    if (!c) break;
	    c++;
	}
	*d=0;
	len += strlen(sl->lines[i]);
    }
    if (!len) return NULL;
    len += 4*i;
    cs=g_malloc(len);
    *cs=0;
    for (i=0;i<8 && sl->lines[i];i++) {
	if (i) strcat(cs," ");
	strcat(cs,sl->lines[i]);
    }
    line=g_malloc0(sizeof(*line));
    if (unit_frames) {
	line->frames=sl->t1;
    }
    else {
	line->decisecs=sl->t1/100;
    }
    line->maxlen=-1;
    line->flags=0;
    line->origin_text=cs;
    line->ortofn_text=NULL;
    line->wave=NULL;
    line->wave_length=0;
    return line;
}
    
    



static int get_hms(char **str)
{
	int h,m,s;
	while (**str && isspace(**str)) (*str)++;
	if (!isdigit(**str)) return -1;
	h=strtol(*str,str,10);
	if (*(*str)++ != ':') return -1;
	if (!isdigit(**str)) return -1;
	m=strtol(*str,str,10);
	if (m<0 || m>59) return -1;
	if (*(*str)++ != ':') return -1;
	if (!isdigit(**str)) return -1;
	s=strtol(*str,str,10);
	if (s<0 || s>59) return -1;
	return 3600 * h + 60 * m + s;
}

char *sub_getline(char **str)
{
	char *c,*d,z;
	if (!*str || !**str) return 0;
	c=*str;
	d=strpbrk(c,"\r\n");
	if (!d) {
		d=c+strlen(c);
	}
	else {
		z=*d;
		*d++=0;
		if (z == '\r' && *d=='\n') d++;
	}
	*str=d;
	return c;
}

static void split_subline(struct subline *line,char *sep)
{
	char *c=line->lines[0];
	int n=1;
	while (n<7) {
		c=strstr(c,sep);
		if (!c) break;
		*c=0;
		c+=strlen(sep);
		line->lines[n++]=c;
	}
}


static struct subline *subr_micro(char *s,int *ifps)
{
	struct subline *slines,**sline,*sl;
	char *line;
	double fps;
	int t1,t2,wfl,fps_set;
	slines=NULL;
	sline=&slines;
	wfl=0;
	fps_set=0;
	if (use_fps_set) {
	    fps_set=1;
	    fps=use_fps;
	}
	while ((line=sub_getline(&s))) {
		if (*line != '{') continue;
		line++;
		if (!isdigit(*line)) continue;
		t1=strtol(line,&line,10);
		if (*line++ !='}') continue;
		if (*line++ !='{') continue;
		if (isdigit(*line)) t2=strtol(line,&line,10);
		else t2=-1;
		if (*line++ != '}') continue;
		while (*line && isspace(*line)) line++;
		if (!*line) continue;
		if (!wfl) {
			char *cc;
			double fdx;
			wfl=1;
			if (t1 >= t2) {
				fdx=strtod(line,&cc);
				if (cc > line && fdx > 20.0 && fdx < 60.0) {
					while (*cc && isspace(*cc)) continue;
					if (!*cc) {
						if (!fps_set) {
						    fps=fdx;
						    fps_set=1;
						}
						continue;
					}
				}
			}
		}
		if (fps_set) {
		    t1=(1000.0 * t1) / fps;
		    t2=(1000.0 * t2) / fps;
		}
		sl=alloc_subline(line,t1,t2);
		*sline=sl;
		sline=&(sl->next);
		split_subline(sl,"|");
	}
	for (sl=slines;sl;sl=sl->next) {
		if (sl->t2 < 0) {
			if (sl->next) sl->t2=sl->next->t1;
			else sl->t2=sl->t1+32;
		}
	}
	if (!fps_set) *ifps=1;
	return slines;
}

static struct subline *subr_mpl2(char *s,int *ifps)
{
	struct subline *slines,**sline,*sl;
	char *line;
	int t1,t2;
	slines=NULL;
	sline=&slines;
	while ((line=sub_getline(&s))) {
		if (*line != '[') continue;
		line++;
		if (!isdigit(*line)) continue;
		t1=strtol(line,&line,10)*100;
		if (*line++ !=']') continue;
		if (*line++ !='[') continue;
		if (!isdigit(*line)) continue;
		t2=strtol(line,&line,10)*100;
		if (*line++ != ']') continue;
		while (*line && isspace(*line)) line++;
		if (!*line) continue;
		sl=alloc_subline(line,t1,t2);
		*sline=sl;
		sline=&(sl->next);
		split_subline(sl,"|");
	}
	return slines;
}

static struct subline *subr_tmp(char *s,int *ifps)
{
	struct subline *slines,**sline,*sl;
	char *line;
	int last_msec=0;
	int t1,t2;
	slines=NULL;
	sline=&slines;
	while ((line=sub_getline(&s))) {
		t1=1000*get_hms(&line);
		if (t1<0) continue;
		if (t1 < last_msec+1000) t1+=last_msec+1000;
		last_msec=t1;
		t2=-1;
		if (*line++!=':') continue;
		while (*line && isspace(*line)) line++;
		if (!*line) continue;
		sl=alloc_subline(line,t1,t2);
		*sline=sl;
		sline=&(sl->next);
		split_subline(sl,"|");
	}
	for (sl=slines;sl;sl=sl->next) {
		if (sl->t2 < 0) {
			sl->t2=sl->t1+4000;
			if (sl->next && sl->t2 > sl->next->t1) sl->t2=sl->next->t1;
		}
	}
	return slines;
}

static struct subline *subr_jaco(char *s,int *ifps)
{
	struct subline *slines,**sline,*sl;
	char *line;
	int timeres=100;
	int t1,t2;
	slines=NULL;
	sline=&slines;
	while ((line=sub_getline(&s))) {
		if (*line=='#') {
			line++;
			if (tolower(*line) == 't') {
				while (*line && !isdigit(*line)) line++;
				int n=strtol(line,&line,10);
				if (n<=100 && n >= 12) timeres=n;
			}
			continue;
		}
		t1=get_hms(&line);
		if (t1<0) continue;
		if (*line++!='.') continue;
		t1=t1*100+strtol(line,&line,10) % 100;
		while (*line && isspace(*line)) line++;
		t2=get_hms(&line);
		if (t2<0) continue;
		if (*line++!='.') continue;
		t2=t2*100+strtol(line,&line,10) % 100;
		while (*line && isspace(*line)) line++;
		sl=alloc_subline(line,t1,t2);
		*sline=sl;
		sline=&(sl->next);
		split_subline(sl,"\\n");
	}
	for (sl=slines;sl && timeres < 100;sl=sl->next) {
		if (sl->t1 % 100 >= timeres || sl->t2 % 100 >= timeres) {
			timeres=100;
			break;
		}
	}
	for (sl=slines;sl;sl=sl->next) {
		sl->t1 = ((sl->t1 /100) * 1000) + ((sl->t1 % 100) * 10)/timeres;
		sl->t2 = ((sl->t2 /100) * 1000) + ((sl->t2 % 100) * 10)/timeres;
	}
	return slines;
}

static struct subline *subr_srip(char *s,int *ifps)
{
	struct subline *slines,**sline,*sl;
	char *line;
	GString *gs;
	int t1,t2;
	slines=NULL;
	sline=&slines;
	while ((line=sub_getline(&s))) {
		if (!*line) continue;
		if (!isdigit(*line)) continue;
		while (*line && isdigit(*line)) line++;
		while (*line && isspace(*line)) line++;
		if (!*line) {
			line=sub_getline(&s);
			if (!line) break;
		}
		t1=get_hms(&line);
		if (t1<0) continue;
		if (*line++ != ',') continue;
		t1=t1*1000+strtol(line,&line,10);
		while (*line && isspace(*line)) line++;
		if (strncmp(line,"-->",3)) continue;
		line+=3;
		while (*line && isspace(*line)) line++;
		t2=get_hms(&line);
		if (t2<0) continue;
		if (*line++ != ',') continue;
		t2=t2*1000+strtol(line,&line,10);
		if (!(line=sub_getline(&s))) break;
		if (!*line) continue;
		gs=g_string_new(line);
		for (;*s;) {
			line=sub_getline(&s);
			if (!line || !*line) break;
			g_string_append_printf(gs,"\n%s",line);
		}
		sl=alloc_subline(NULL,t1,t2);
		sl->lines[0]=g_string_free(gs,0);
		*sline=sl;
		sline=&(sl->next);
		split_subline(sl,"\n");
	}
	return slines;
}

static struct subline *subr_subv(char *s,int *ifps)
{
	struct subline *slines,**sline,*sl;
	char *line;
	int t1,t2;
	GString *gs;
	slines=NULL;
	sline=&slines;
	while (*s && (*s =='\r' || *s=='\n' || *s=='[')) sub_getline(&s);
	while ((line=sub_getline(&s))) {
		t1=get_hms(&line);if (t1<0) continue;
		if (*line++ !='.') continue;
		t1=t1*1000 + strtol(line,&line,10) * 10;
		if (*line++!=',') continue;
		t2=get_hms(&line);if (t2<0) continue;
		if (*line++ !='.') continue;
		t2=t2*1000 + strtol(line,&line,10) * 10;
		if (!(line=sub_getline(&s))) break;
		if (!*line) continue;
		gs=g_string_new(line);
		for (;*s;) {
			line=sub_getline(&s);
			if (!line || !*line) break;
			g_string_append_printf(gs,"[br]%s",line);
		}
		sl=alloc_subline(NULL,t1,t2);
		sl->lines[0]=g_string_free(gs,0);
		*sline=sl;
		sline=&(sl->next);
		split_subline(sl,"[br]");
	}
	return slines;
}

static struct subline *subr_ass(char *s,int *ifps)
{
	struct subline *slines,**sline,*sl;
	char *line;
	int t1,t2;
	int imode=0;
	int format=0,tp1=-1,tp2=-1,step=0;
	char *c,*d;
	
	void debra(char *c)
	{
	    char *d=c;
	    int rs=0;
	    for (;*c;c++) {
		if (*c=='{') {
		    rs=1;
		    continue;
		}
		if (*c=='}') {
		    rs=0;
		    continue;
		}
		if (!rs) *d++=*c;
	    }
	    *d=0;
	}
	
	slines=NULL;
	sline=&slines;
	while ((line=sub_getline(&s))) {
	    if (line[0] == '[') {
		imode=(!strncasecmp(line,"[events]",8))?1:0;
		continue;
	    }
	    if (!imode) continue;
	    if (!strncasecmp(line,"format:",7)) {
		for (c=line,format=0,tp1=tp2=-1;*c;) {
		    while (*c && isspace(*c)) c++;
		    if (!*c) break;
		    if (!strncasecmp(c,"start",4)) tp1=format;
		    else if (!strncasecmp(c,"end",3)) tp2=format;
		    c=strchr(c,',');
		    if (!c) break;
		    c++;
		    format++;
		}
		imode=2;
		continue;
	    }
	    if (!strncasecmp(line,"dialogue:",9) && imode == 2) {
		line+=9;
		while (*line && isspace(*line)) line++;
		debra(line);
		step=0;
		while (line && *line && step < format) {
		    if (step == tp1) {
			t1=get_hms(&line) * 1000;
			if (*line++=='.') {
			    t1+=strtol(line,&line,10)*10;
			}
			else break;
		    }
		    else if (step == tp2) {
			t2=get_hms(&line) * 1000;
			if (*line++=='.') {
			    t2+=strtol(line,&line,10)*10;
			}
			else break;
		    }
		    line=strchr(line,',');
		    if (!line) break;
		    line++;
		    step++;
		}
		if (step < format) continue;
		for (c=d=line;*c;c++) {
		    if (*c=='\\' && (c[1]=='N' || c[1]=='n')) {
			*d++=' ';
			c++;
		    }
		    else {
			*d++=*c;
		    }
		}
		*d=0;
		if (tp2 < 0) t2=t1;
		sl=alloc_subline(line,t1,t2);
		for (sline=&slines;*sline;sline=&(*sline)->next) {
		    if ((*sline)->t1 > t1) break;
		}
		sl->next=*sline;
		*sline=sl;
		continue;
	    }
	}
	return slines;
}

// testery
// MPL2: [d][d]
// Micro: {d}{D}
// TMP: 99:59:59:
// SubRip: Ed:59:59,d --> d:59:59,d
// SubViewer: d:S:S:99,d:S:S:99
// JacoSUB: d:S:S.99 d:S:S.99

static struct subtester {
	char *tester;
	int flags;
	struct subline *(*fun)(char *,int *);
	char *name;
} substruct[]={
	{"[d][d]",0,subr_mpl2,"mpl2"},
	{"{d}{D}",1,subr_micro,"micro"},
	{"Ed:59:59,d --> d:59:59,d",0,subr_srip,"srip"},
	{"d:S:D.99,d:S:S.99",0,subr_subv,"subv"},
	{"99:59:59:",0,subr_tmp,"tmp"},
	{"d:S:S.99 d:S:S.99",2,subr_jaco,"jaco"},
	{NULL,0,NULL}},
	ass={"",0,subr_ass,"ass"};



static const char *get_encoding(char *buf,size_t len,char *lang)
{
	EncaAnalyser an;char lb[4];
	const double mu = 0.005;
	const double m = 15.0;
	size_t sgnf;
	const char *c;
	EncaEncoding result;
	for (c=buf;*c && (c-buf)<len;c++) if ((*c) & 0x80) break;
	if (!*c || c>=buf+len) return "UTF-8";
	if (!lang) {
	    c=getenv("LANG");
	    if (c) {
		strncpy(lb,c,2);
		lb[2]=0;
		lang=lb;
	    }
	    else {
		lang="en";
	    }
	}
	an=enca_analyser_alloc(lang);
	if (!an) {
		return NULL;
	}
	enca_set_threshold(an, 1.38);
	enca_set_multibyte(an, 1);
	enca_set_ambiguity(an, 1);
	enca_set_garbage_test(an, 1);
	sgnf = ceil((double)len/(len/m + 1.0/mu));
	enca_set_significant(an, sgnf);
	enca_set_termination_strictness(an, 1);
	enca_set_filtering(an, sgnf > 2);

	result = enca_analyse_const(an, (unsigned char *)buf, len);
	if (!enca_charset_is_known(result.charset)) {
		enca_analyser_free(an);
		return NULL;
	}
	c=enca_charset_name(result.charset,ENCA_NAME_STYLE_ICONV);
	enca_analyser_free(an);
	return c;
}

static struct subtitle *_subtitle_read_buffer(char *fname,char *encoding,char *lang,int *is_frames)
{
	struct subtitle *sbt,**ssbt;
	char *fbuf=NULL;
	gsize flen;
	char *c;int n,i;
	struct subtester *ss;
	int ifps=0;
	struct subline *lines,*li;
	if (!fname) return NULL;
	fprintf(stderr,"Getting file [%s]\n",fname);
	if (!g_file_get_contents(fname,&fbuf,&flen,NULL)) return NULL;
	if (!encoding) {
	    encoding=(char *)get_encoding(fbuf,flen,lang);
	}
	if (!encoding) encoding="CP1250";
	fprintf(stderr,"Encoding found %s\n",encoding);
	if (strcasecmp(encoding,"utf8") && strcasecmp(encoding,"utf-8")) {
		c=fbuf;
		fbuf=g_convert(c,flen,"UTF-8",encoding,NULL,NULL,NULL);
		g_free(c);
		
	}
	for (c=fbuf,n=0,ss=NULL;c && n<100 && !ss;n++,c=strchr(c,'\n')) {
		for (i=0;substruct[i].tester;i++) {
			if (tester_pasuje(c,substruct[i].tester)) {
				fprintf(stderr,"Subtitle format found: %s\n",substruct[i].name);
				ss=&substruct[i];
				break;
			}
		}
	}
	if (!ss) {
	    c=fbuf;
	    if (!memcmp(c,"\xef\xbb\xbf",3)) {
		c+= 3;
	    }
	    if (!strncasecmp(c,"[Script Info]",13)) {
		fprintf(stderr,"Subtitle format found: %s\n",ass.name);
		ss=&ass;
	    }
	}
	if (!ss) {
		fprintf(stderr,"Subtitle format not found\n");
		g_free(fbuf);
		return NULL;
	}
	ifps=0;
	lines=ss->fun(fbuf,&ifps);
	g_free(fbuf);
	sbt=NULL;
	ssbt=&sbt;
	while ((li=lines)) {
		struct subtitle *su;
		lines=lines->next;
		su=prepare_line(li,ifps);
		if (su) {
		    *ssbt=su;
		    ssbt=&(su->next);
		}
		g_free(li->lines[0]);
		g_free(li);
	}
	if (is_frames) *is_frames=ifps;
	return sbt;
}

struct subtitle *subtitle_read_buffer(char *fname,char *movie_name,char *encoding,char *lang,int *is_frames)
{
	static char *exts[]={".asub",".sub",".srt",".txt",".ass",".ssa",NULL};
	struct subtitle *rc=NULL;
	char *c;
	int i;
	if (fname) return _subtitle_read_buffer(fname,encoding,lang,is_frames);
	fname=g_malloc(strlen(movie_name)+10);
	strcpy(fname,movie_name);
	c=strrchr(fname,'/');
	if (!c) c=fname;
	c=strrchr(c,'.');
	if (!c) c=fname+strlen(fname);
	for (i=0;exts[i];i++) {
		strcpy(c,exts[i]);
		rc=_subtitle_read_buffer(fname,encoding,lang,is_frames);
		if (rc) break;
	}
	g_free(fname);
	return rc;
}

void make_subarray(void)
{
    struct subtitle *s;int n;
    for (subarray_len=0,s=sublist;s;s=s->next,subarray_len++);
    subarray=g_malloc(sizeof(*subarray) * subarray_len);
    for (s=sublist,n=0;s;s=s->next,n++) subarray[n]=s;
}
